{------------------------------------------------------------------------------}
{ Projet: Compression et dcompression d'archives.                             }
{ Objet : Unit base sur LZSS + Huffman coding.                               }
{ Auteur: J.C. Vandenhekke.                                                    }
{------------------------------------------------------------------------------}
{ LZHUF.C English version 1.0                                                  }
{ Based on Japanese version 29-NOV-1988 LZSS coded by Haruhiko OKUMURA         }
{ Adaptive Huffman Coding coded by Haruyasu YOSHIZAKI                          }
{ Edited and translated to English by Kenji RIKITAKE                           }
{ Translated from C to Turbo Pascal by Douglas Webb   2/18/91                  }
{    Update and bug correction of TP version 4/29/91 (Sorry!!)                 }
{                                                                              }
{   This Unit allows the user to commpress data using a combination of         }
{ LZSS Compression and adaptive Huffman coding, or conversely to deCompress    }
{ data that was previously Compressed by this Unit.                            }
{                                                                              }
{ NOTE: All the large data structures For these routines are allocated when    }
{ needed from the heap, and deallocated when finished.  So when not in use     }
{ memory requirements are minimal.  However, this Unit Uses about 34K of       }
{ heap space, and 400 Bytes of stack when in use.                              }
{                                                                              }
{ 23/03/1997 : v1.00 - Utilisation et modifications par JC V.                  }
{------------------------------------------------------------------------------}
{ 21/12/1997 : Dernire modification.                                          }
{------------------------------------------------------------------------------}

{$A+,R-} {   $A+,B-,D-,E-,F-,I+,L-,N-,O-,R-,S-,V-}

UNIT Lzhuf;

INTERFACE

USES Sysutils;

TYPE
{$IFDEF WIN32}
   Int16 = SmallInt;
{$ELSE}
   Int16 = Integer;
{$ENDIF}

  ElzhException = CLASS (Exception);

  PutOneByteProc = FUNCTION (aByte : Byte; Flush : Boolean) : Boolean;
  GetOneByteProc = FUNCTION (VAR aByte : Byte) : Boolean;

PROCEDURE LZHPack (VAR BytesWritten : LongInt;
                       GetOneByte   : GetOneByteProc;
                       PutOneByte   : PutOneByteProc);

PROCEDURE LZHUnpack (TextSize   : LongInt;
                     GetOneByte : GetOneByteProc;
                     PutOneByte : PutOneByteProc);

{==============================================================================}

IMPLEMENTATION

CONST  Exit_OK     = 0;
       Exit_FAILED = 1;

       { LZSS Parameters }
       N	   = 4096;	{ Size of string buffer }
       F	   = 60;	{ Size of look-ahead buffer }
       THRESHOLD   = 2;
       NUL         = N;   	{ End of tree's node  }

       { Huffman coding parameters }
       N_CHAR = (256 - THRESHOLD + F);

       { character code (:= 0..N_CHAR-1) }
       T = (N_CHAR * 2 - 1);	 { Size of table }
       R = (T - 1);		 { root position }

       { update when cumulative frequency reaches to this value }
       MAX_FREQ = $8000;

{ Tables FOR encoding/decoding upper 6 bits of sliding dictionary pointer }

{--- encoder table ------------------------------------------------------------}

  p_len : ARRAY [0..63] OF BYTE =
  ($03, $04, $04, $04, $05, $05, $05, $05, $05, $05, $05, $05, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06, $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08);

  p_code : ARRAY [0..63] OF BYTE =
  ($00, $20, $30, $40, $50, $58, $60, $68, $70, $78, $80, $88, $90, $94, $98, $9C,
   $A0, $A4, $A8, $AC, $B0, $B4, $B8, $BC, $C0, $C2, $C4, $C6, $C8, $CA, $CC, $CE,
   $D0, $D2, $D4, $D6, $D8, $DA, $DC, $DE, $E0, $E2, $E4, $E6, $E8, $EA, $EC, $EE,
   $F0, $F1, $F2, $F3, $F4, $F5, $F6, $F7, $F8, $F9, $FA, $FB, $FC, $FD, $FE, $FF);

{--- decoder table ------------------------------------------------------------}

  d_code : ARRAY [0..255] OF BYTE =
  ($00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00,
   $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01, $01,
   $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02, $02,
   $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04, $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06, $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08, $09, $09, $09, $09, $09, $09, $09, $09,
   $0A, $0A, $0A, $0A, $0A, $0A, $0A, $0A, $0B, $0B, $0B, $0B, $0B, $0B, $0B, $0B,
   $0C, $0C, $0C, $0C, $0D, $0D, $0D, $0D, $0E, $0E, $0E, $0E, $0F, $0F, $0F, $0F,
   $10, $10, $10, $10, $11, $11, $11, $11, $12, $12, $12, $12, $13, $13, $13, $13,
   $14, $14, $14, $14, $15, $15, $15, $15, $16, $16, $16, $16, $17, $17, $17, $17,
   $18, $18, $19, $19, $1A, $1A, $1B, $1B, $1C, $1C, $1D, $1D, $1E, $1E, $1F, $1F,
   $20, $20, $21, $21, $22, $22, $23, $23, $24, $24, $25, $25, $26, $26, $27, $27,
   $28, $28, $29, $29, $2A, $2A, $2B, $2B, $2C, $2C, $2D, $2D, $2E, $2E, $2F, $2F,
   $30, $31, $32, $33, $34, $35, $36, $37, $38, $39, $3A, $3B, $3C, $3D, $3E, $3F);

  d_len : ARRAY [0..255] OF BYTE =
  ($03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04, $04,
   $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08, $08);

  getbuf         : Word    = 0;
  getlen         : Byte    = 0;
  putlen         : Byte    = 0;
  putbuf         : Word    = 0;
  TextSize       : LongInt = 0;
  codesize       : LongInt = 0;
  printcount     : LongInt = 0;
  match_position : Int16   = 0;
  match_length   : Int16   = 0;

TYPE
  FreqType    = ARRAY [0..T] OF Word;
  FreqPtr     = ^freqtype;
  PntrType    = ARRAY [0..PRED (T + N_Char)] OF Int16;
  pntrPtr     = ^pntrType;
  SonType     = ARRAY [0..PRED (T)] OF Int16;
  SonPtr      = ^SonType;
  TextBufType = ARRAY [0..N + F - 2] OF Byte;
  TBufPtr     = ^TextBufType;
  WordRay     = ARRAY [0..N] OF Int16;
  WordRayPtr  = ^WordRay;
  BWordRay    = ARRAY [0..N + 256] OF Int16;
  BWordRayPtr = ^BWordRay;

VAR
  text_buf  : TBufPtr;
  lson, dad : WordRayPtr;
  rson      : BWordRayPtr;
  freq      : FreqPtr;	    { cumulative freq table }

{ * pointing parent nodes.
  * area [T..(T + N_CHAR - 1)] are pointers FOR leaves }
  prnt : PntrPtr;

{ pointing children nodes (son[], son[] + 1)}
  son : SonPtr;

{--- Initializing tree --------------------------------------------------------}

PROCEDURE InitTree;

  VAR i : Int16;

  BEGIN
    FOR i := N + 1 TO N + 256  DO rson^[i] := NUL;	{ root }
    FOR i := 0 TO N DO dad^[i] := NUL;			{ node }
  END; { InitTree }

{--- Inserting node to the tree -----------------------------------------------}

PROCEDURE InsertNode (r : Int16);

  VAR tmp, i, p, cmp : Int16;
      key : TBufPtr;
      c : WORD;

  BEGIN
    cmp := 1;
    key := @text_buf^[r];
    p := SUCC(N) + key^[0];
    rson^[r] := NUL;
    lson^[r] := NUL;
    match_length := 0;
    WHILE match_length < F DO
      BEGIN
        IF (cmp >= 0) THEN
          BEGIN
            IF (rson^[p] <> NUL)
            THEN p := rson^[p]
            ELSE
              BEGIN
                rson^[p] := r;
                dad^[r] := p;
		exit;
              END;
          END
        ELSE
          BEGIN
            IF (lson^[p] <> NUL)
            THEN p := lson^[p]
            ELSE
              BEGIN
                lson^[p] := r;
	        dad^[r] := p;
		exit;
              END;
          END;
        i := 0;
        cmp := 0;
        WHILE (i < F) AND (cmp = 0) DO
          BEGIN
            inc (i);
            cmp := key^[i] - text_buf^[p + i];
          END;
        IF (i > THRESHOLD) THEN
          BEGIN
            tmp := PRED ((r - p) AND PRED (N));
	    IF (i > match_length) THEN
              BEGIN
                match_position := tmp;
                match_length := i;
              END;
            IF (match_length < F) AND (i = match_length) THEN
              BEGIN
                c := tmp;
	        IF (c < Word (match_position))
                  THEN match_position := c;
              END;
          END; { if i > threshold }
      END; { WHILE match_length < F }
    dad^[r] := dad^[p];
    lson^[r] := lson^[p];
    rson^[r] := rson^[p];
    dad^[lson^[p]] := r;
    dad^[rson^[p]] := r;
    IF (rson^[dad^[p]] = p)
      THEN rson^[dad^[p]] := r
      ELSE lson^[dad^[p]] := r;
    dad^[p] := NUL;  { remove p }
  END; { InsertNode }

{--- Deleting node from the tree ----------------------------------------------}

PROCEDURE DeleteNode (p: Int16);

  VAR q : Int16;

  BEGIN
    IF (dad^[p] = NUL) THEN exit; { unregistered }
    IF (rson^[p] = NUL)
    THEN q := lson^[p]
    ELSE
      BEGIN
        IF (lson^[p] = NUL)
        THEN q := rson^[p]
        ELSE
          BEGIN
            q := lson^[p];
            IF (rson^[q] <> NUL) THEN
              BEGIN
                REPEAT
                  q := rson^[q];
                UNTIL (rson^[q] = NUL);
                rson^[dad^[q]] := lson^[q];
                dad^[lson^[q]] := dad^[q];
                lson^[q] := lson^[p];
                dad^[lson^[p]] := q;
              END;
            rson^[q] := rson^[p];
            dad^[rson^[p]] := q;
          END;
      END;
    dad^[q] := dad^[p];
    IF (rson^[dad^[p]] = p)
      THEN rson^[dad^[p]] := q
      ELSE lson^[dad^[p]] := q;
    dad^[p] := NUL;
  END; { DeleteNode }

{--- Huffman coding parameters ------------------------------------------------}

FUNCTION GetBit (GetOneByte : GetOneByteProc) : Int16; { get one bit }

  VAR i  : Byte;
      i2 : Int16;

  BEGIN
    WHILE (getlen <= 8)
    DO BEGIN
         IF GetOneByte (i)
           THEN i2 := i
           ELSE i2 := 0;
         getbuf := getbuf OR (Word (i2) SHL (8 - getlen));
         inc (getlen, 8);
       END;
    i2 := Int16 (getbuf);
    getbuf := getbuf SHL 1;
    dec (getlen);
    GetBit := Int16 (i2 < 0);
  END; { GetBit }

FUNCTION GetByte (GetOneByte : GetOneByteProc) : Int16; { get a Byte }

  VAR j : Byte;
      i : Word;

  BEGIN
    WHILE (getlen <= 8)
    DO BEGIN
         IF GetOneByte (j)
           THEN i := j
           ELSE i := 0;
         getbuf := getbuf OR (i SHL (8 - getlen));
         inc (getlen, 8);
       END;
    i := getbuf;
    getbuf := getbuf SHL 8;
    dec (getlen, 8);
    GetByte := Int16 (i SHR 8);
  END; { GetByte }

PROCEDURE Putcode (l : Int16; c : WORD; PutOneByte : PutOneByteProc);		{ output c bits }

  VAR Temp : BYTE;

  BEGIN
    putbuf := putbuf OR (c SHR putlen);
    inc (putlen, l);
    IF (putlen >= 8) THEN
      BEGIN
        Temp := putbuf SHR 8;
        PutOneByte (Temp, false);
        DEC (putlen, 8);
        IF (putlen  >= 8) THEN
          BEGIN
            Temp := Lo (PutBuf);
            PutOneByte (Temp, false);
            INC (codesize, 2);
            DEC (putlen, 8);
            putbuf := c SHL (l - putlen);
          END
        ELSE
          BEGIN
	    putbuf := putbuf SHL 8;
	    INC (codesize);
          END;
      END;
  END; { Putcode }

{--- initialize freq tree -----------------------------------------------------}

PROCEDURE StartHuff;

  VAR i, j : Integer;

  BEGIN
    FOR i := 0 TO pred (N_Char)
    DO BEGIN
         freq^[i] := 1;
         son^[i] := i + T;
         prnt^[i + T] := i;
       END;
    i := 0;
    j := N_Char;
    WHILE (j <= R)
    DO BEGIN
         freq^[j] := freq^[i] + freq^[i + 1];
         son^[j] := i;
         prnt^[i] := j;
         prnt^[i + 1] := j;
         inc(i, 2);
         inc(j);
       END;
    freq^[T] := $ffff;
    prnt^[R] := 0;
  END; { StartHuff }

{--- reConstruct freq tree ----------------------------------------------------}

PROCEDURE reConst;

  VAR i, j, k, tmp : Integer;
      F, l : Word;

  BEGIN
    { halven cumulative freq For leaf nodes }
    j := 0;
    FOR i := 0 TO pred (T)
    DO BEGIN
         IF (son^[i] >= T)
         THEN BEGIN
                freq^[j] := succ (freq^[i]) DIV 2; {@@ Bug Fix MOD -> div @@}
                son^[j] := son^[i];
                inc(j);
              END;
       END;
    { make a tree : first, connect children nodes }
    i := 0;
    j := N_Char;
    WHILE (j < T)
    DO BEGIN
         k := succ (i);
         F := freq^[i] + freq^[k];
         freq^[j] := F;
         k := pred(j);
         WHILE F < freq^[k] DO dec (k);
         inc (k);
         l := (j - k) SHL 1;
         tmp := succ (k);
         move (freq^[k], freq^[tmp], l);
         freq^[k] := F;
         move (son^[k], son^[tmp], l);
         son^[k] := i;
         inc (i, 2);
         inc (j);
       END;
    { connect parent nodes }
    FOR i := 0 TO pred (T)
    DO BEGIN
         k := son^[i];
         IF (k >= T)
         THEN BEGIN
                prnt^[k] := i;
              END
         ELSE BEGIN
                prnt^[k] := i;
                prnt^[succ(k)] := i;
              END;
       END;
  END; { reConst }

{--- update freq tree ---------------------------------------------------------}

PROCEDURE update (c : Integer);

  VAR i, j, k, l : Integer;

  BEGIN
    IF (freq^[R] = MAX_FREQ)
    THEN BEGIN
           reConst;
         END;
    c := prnt^[c + T];
    REPEAT
      inc (freq^[c]);
      k := freq^[c];
      { swap nodes to keep the tree freq-ordered }
      l := succ (c);
      IF (k > freq^[l])
      THEN BEGIN
             WHILE (k > freq^[l]) DO inc (l);
             dec (l);
             freq^[c] := freq^[l];
             freq^[l] := k;
             i := son^[c];
             prnt^[i] := l;
             IF (i < T) THEN prnt^[succ (i)] := l;
             j := son^[l];
             son^[l] := i;
             prnt^[j] := c;
             IF (j < T) THEN prnt^[succ (j)] := c;
             son^[c] := j;
             c := l;
           END;
      c := prnt^[c];
    UNTIL (c = 0);                     { Repeat it Until reaching the root }
  END; { update }

VAR code, len : WORD;

PROCEDURE EncodeChar (c : WORD; PutOneByte : PutOneByteProc);

  VAR i : WORD;
      j, k : Int16;

  BEGIN
    i := 0;
    j := 0;
    k := prnt^[c + T];
    { search connections from leaf node to the root }
    REPEAT
      i := i SHR 1;
      { IF node's address is odd, output 1 ELSE output 0 }
      IF BOOLEAN (k AND 1) THEN INC (i, $8000);
      INC (j);
      k := prnt^[k];
    UNTIL (k = R);
    Putcode (j, i, PutOneByte);
    code := i;
    len := j;
    update (c);
  END; { EncodeChar }

PROCEDURE EncodePosition (c : WORD; PutOneByte : PutOneByteProc);

  VAR i, j : WORD;

  BEGIN
    { output upper 6 bits with encoding }
    i := c SHR 6;
    j := p_code[i];
    Putcode (p_len[i], j SHL 8, PutOneByte);
    { output lower 6 bits directly }
    Putcode  (6, (c AND $3f) SHL 10, PutOneByte);
  END; { EncodePosition }

PROCEDURE EncodeEnd (PutOneByte : PutOneByteProc);

  VAR Temp : BYTE;

  BEGIN
    IF BOOLEAN (putlen) THEN
      BEGIN
        Temp := Lo (putbuf SHR 8);
        PutOneByte (Temp, false);
        INC (codesize);
      END;
  END; { EncodeEnd }

FUNCTION DecodeChar (GetOneByte : GetOneByteProc) : Integer;

  VAR c : Word;

  BEGIN
    c := son^[R];
    { * start searching tree from the root to leaves.
      * choose node #(son[]) if input bit = 0
      * else choose #(son[]+1) (input bit = 1) }
    WHILE (c < T)
    DO BEGIN
         c := c + GetBit (GetOneByte);
         c := son^[c];
       END;
    c := c - T;
    update (c);
    DecodeChar := Integer (c);
  END; { DecodeChar }

FUNCTION DecodePosition (GetOneByte : GetOneByteProc) : Word;

  VAR i, j, c : Word;

  BEGIN
    { decode upper 6 bits from given table }
    i := GetByte (GetOneByte);
    c := Word (d_code[i] SHL 6);
    j := d_len[i];
    { input lower 6 bits directly }
    dec(j, 2);
    WHILE j <> 0
    DO BEGIN
         i := (i SHL 1) + GetBit (GetOneByte);
         dec (j);
       END;
    DecodePosition := c OR i AND $3f;
  END; { DecodePosition }

{--- Initialisation -----------------------------------------------------------}

PROCEDURE InitLZH;

  BEGIN
    getbuf := 0;
    getlen := 0;
    putlen := 0;
    putbuf := 0;
    TextSize := 0;
    codesize := 0;
    printcount := 0;
    match_position := 0;
    match_length := 0;
    TRY
      new (lson);
      new (dad);
      new (rson);
      new (Text_buf);
      new (freq);
      new (prnt);
      new (son);
    EXCEPT
      RAISE ElzhException.Create ('LZH : Cannot get memory for dictionary tables');
    END;
  END; { InitLZH }

{--- Done ---------------------------------------------------------------------}

PROCEDURE endLZH;

  BEGIN
    TRY
      dispose (son);
      dispose (prnt);
      dispose (freq);
      dispose (Text_buf);
      dispose (rson);
      dispose (dad);
      dispose (lson);
    EXCEPT
      RAISE ElzhException.Create ('LZH : Error freeing memory for dictionary tables');
    END;
  END; { endLZH }

{--- Compression --------------------------------------------------------------}

PROCEDURE LZHPack (VAR BytesWritten : LongInt;
                       GetOneByte   : GetOneByteProc;
                       PutOneByte   : PutOneByteProc);

  VAR ct : BYTE;
      i, len, r, s, last_match_length : Int16;
      Got : Boolean;

  BEGIN
    InitLZH;
    try
      textsize := 0;			{ rewind and rescan }
      StartHuff;
      InitTree;
      s := 0;
      r := N - F;
      FillChar (Text_buf^[0], r, ' ');
      len := 0;
      Got := true;
      WHILE (len < F) AND (Got) DO
        BEGIN
          Got := GetOneByte (ct);
          IF Got THEN
            BEGIN
              text_buf^[r + len] := ct;
              INC(len);
            END;
        END;
      textsize := len;
      FOR i := 1 to F DO InsertNode (r - i);
      InsertNode(r);
      REPEAT
        IF (match_length > len) THEN match_length := len;
        IF (match_length <= THRESHOLD) THEN
          BEGIN
            match_length := 1;
	    EncodeChar(text_buf^[r],PutOneByte);
          END
        ELSE
          BEGIN
            EncodeChar(255 - THRESHOLD + match_length,PutOneByte);
	    EncodePosition(match_position,PutOneByte);
          END;
        last_match_length := match_length;
        i := 0;
        Got := true;
        WHILE (i < last_match_length) AND (Got) DO
          BEGIN
            Got := GetOneByte (ct);
            IF Got THEN
              BEGIN
                DeleteNode (s);
                text_buf^[s] := ct;
                IF (s < PRED (F)) THEN text_buf^[s + N] := ct;
                s := SUCC (s) AND PRED (N);
                r := SUCC (r) AND PRED (N);
                InsertNode (r);
                inc (i);
              END;
          END; { endwhile }
        INC (textsize, i);
        WHILE (i < last_match_length) DO
          BEGIN
            INC (i);
            DeleteNode (s);
            s := SUCC (s) AND PRED (N);
            r := SUCC (r) AND PRED (N);
            DEC (len);
            IF BOOLEAN (len) THEN InsertNode (r);
          END; { endwhile }
      UNTIL (len <= 0);  { end repeat }
      EncodeEnd (PutOneByte);
      PutOneByte (0, True); { Flush }
    finally
      EndLZH;
    end;
    BytesWritten := TextSize;
  END; { LZHPack }

{--- Decompression ------------------------------------------------------------}

PROCEDURE LZHUnpack (TextSize : LongInt;
                     GetOneByte : GetOneByteProc;
                     PutOneByte : PutOneByteProc);

  VAR c, i, j, k, R : Integer;
      c2            : Byte;
      count         : LongInt;

  BEGIN
    InitLZH;
    TRY
      StartHuff;
      R := N - F;
      fillChar (Text_buf^[0], R, ' ');
      count := 0;
      WHILE count < TextSize
      DO BEGIN
           c := DecodeChar (GetOneByte);
           IF (c < 256)
           THEN BEGIN
                  c2 := lo(c);
                  PutOneByte (c2, False);
                  Text_buf^[R] := c;
                  inc(R);
                  R := R AND pred (N);
                  inc (count);
                END
           ELSE BEGIN
                  i := (R - succ (DecodePosition (GetOneByte))) AND pred (N);
                  j := c - 255 + THRESHOLD;
                  FOR k := 0 TO pred (j)
                  DO BEGIN
                       c := Text_buf^[(i + k) AND pred (N)];
                       c2 := lo (c);
                       PutOneByte (c2, False);
                       Text_buf^[R] := c;
                       inc (R);
                       R := R AND pred (N);
                       inc (count);
                     END;
                END;
         END;
      PutOneByte (0, True); { Flush }
    FINALLY
      endLZH;
    END;
  END; { LZHUnpack }

END.
