(* GNU-GPL
 * System Compression - Logiciel de Compression/Decompression
 * Copyright (c) 2002 Ulrich Van Den Hekke (ulrich.vdh@free.fr)
 * 24 rue Jean Jaures
 * 59560 Comines
 *
 * Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon
 * les termes de la Licence Publique Gnrale GNU publie par la
 * Free Software Foundation (version 2 ou bien toute autre version ultrieure
 * choisie par vous).
 *
 * Ce programme est distribu car potentiellement utile, mais
 * SANS AUCUNE GARANTIE, ni explicite ni implicite, y compris les garanties de
 * commercialisation ou d'adaptation dans un but spcifique. Reportez-vous  la
 * Licence Publique Gnrale GNU pour plus de dtails.
 *
 * Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme
 * temps que ce programme ; si ce n'est pas le cas, crivez  la
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis.
 *
 *)

(******************************************************************************)
(* Structure de l'archive                                                     *)
(* Version : 1.0.0.0                                                          *)
(* Author : Ulrich Van Den Hekke                                              *)
(* Create day : 06/07/2002                                                    *)
(******************************************************************************)

unit scFileStruct;

interface

uses Windows, Classes, SysUtils, Common, AlgoList, VStream;

type
  ESystemCFileStruct = class(Exception);

  {$IFDEF CONSOLEAPP}
  TOnAskNewDisk      = function (const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean;
  TOnAskOneDisk      = function (const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean;
  {$ELSE}
  TOnAskNewDisk      = function (const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean of object;
  TOnAskOneDisk      = function (const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean of object;
  {$ENDIF}

  TSystemCFileStruct = class
  private
    FList: TList;
    FAlgo: TAlgoPluginsList;

    FNeedFlush     : boolean;

    FCurrentStream : TStream;
    FCurrentMaxSize: LongInt;
    FCurrentDisk   : Byte;
    FStartOfList   : TPosition;
    FLastDisk      : Byte;

    FOnAskNewDisk: TOnAskNewDisk;
    FOnAskOneDisk: TOnAskOneDisk;
    FOnProgress: TOnProgression;

    function GetCount(): Integer;
    function GetItems(Index: Integer): PFileInformation;
    procedure SetOnAskNewDisk(const Value: TOnAskNewDisk);
    procedure SetOnAskOneDisk(const Value: TOnAskOneDisk);

    function OnWriteStream(var NewStream: TStream; var NewSpace: LongInt): boolean;
    function OnReadStream(var NewStream: TStream; var NewSpace: LongInt): boolean;
    procedure SetCurrentMaxSize(const Value: LongInt);
    procedure SetOnProgress(const Value: TOnProgression);
  protected
    procedure LoadFileList();
    procedure CreateHeadFile();

    procedure AskNewDisk();
    procedure AskOneDisk(const DiskNum: Byte);

    property CurrentMaxSize: LongInt read FCurrentMaxSize write SetCurrentMaxSize;
  public
    constructor Create(const Archive: TStream; const MaxSize: LongInt; const List: TAlgoPluginsList; const OnNew: TOnAskNewDisk; const OnOne: TOnAskOneDisk);
    destructor Destroy(); override;

    procedure Flush();
    procedure CheckFile();
    function IsMultiDisk: boolean;

    function Add(const RealFileInfo: TRealFileInfo; const input: TStream; const IAlgo: integer = -1): integer; overload;
    function Add(const FileName: TFileName; const IAlgo: integer = -1; const orignalPath: string = ''): integer; overload;
    procedure Delete(index: Integer);
    procedure Extract(Index: Integer; output: TStream); overload;
    procedure Extract(Index: Integer; path: TFileName); overload;

    function IndexOf(FileName: TFileName; const index: integer = 1): integer;

    property Items[Index: Integer]: PFileInformation read GetItems; default;
    property Count: Integer read GetCount;
    property CurrentDisk : byte read FCurrentDisk;
    property NumberOfDisk: byte read FLastDisk;

    property OnAskNewDisk: TOnAskNewDisk read FOnAskNewDisk write SetOnAskNewDisk;
    property OnAskOneDisk: TOnAskOneDisk read FOnAskOneDisk write SetOnAskOneDisk;

    property OnProgress  : TOnProgression read FOnProgress write SetOnProgress;

    property AlgoList    : TAlgoPluginsList read FAlgo;
  end;

  TIdentFile = array[1..6] of char;

const
  cstQMax4Test   = 512 * 1024; (* 512 Ko *)
  cstIdentFile   : TIdentFile = 'SCFILE';
  cstIdentFin    : TIdentFile = 'FIN#00';
  cstMinSize     = sizeof(TIdentFile)       + // Identification du type de fichier
                   sizeof(Byte)             + // Le numero du disque
                   sizeof(TIdentFile)       + // L'entete de fin
                   sizeof(TPosition)        + // La position du debut des entetes;
                   sizeof(Integer)          + // Le nombre d'entetes
                   sizeof(LongWord)         + // Le CRC des entetes
                   sizeof(TFileInformation) + // Un Entete
                   1;                         // Une petite marge
implementation

uses FileTools;

resourcestring
  rsAlgoNotFound    = 'L''algorythme n''a pas t trouv dans la liste des plugins disponibles';
  rsAbort           = 'Opration interrompue';
  rsFileNotEnd      = 'Il faut ouvrir le fichier du dernier disque';
  rsBadFileFormat   = 'Le fichier spcifi est du mauvais format';
  rsNoAssignedEvent = 'Aucun vnement n''a t assign aux mthodes';
  rsMinSize         = 'La taille mimimum d''une archive est de %d octets';
  rsBadCRC          = 'Mauvais CRC lors de la lecture de %s [Vaut %s mais devrait tre %s]';
  rsFileInfos       = 'base des fichiers archivs';

  rsFlush           = 'base de donne';
  rsDelete          = 'Suppression de %d';
  rsLoad            = 'base de donne';

{ TSystemCFileStruct }

function TSystemCFileStruct.Add(const FileName: TFileName; const IAlgo: integer; const orignalPath: string): integer;
var Source: TStream;
    Info  : TRealFileInfo;
    FName,
    FPath : string;
begin
  FillChar(Info, sizeof(Info), #0);
  Source := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    // Fichier Reel
    FName := ExtractFileName(FileName);
    FPath := ExtractFilePath(FileName);
    FPath := Copy(FPath, pos(NormalDir(orignalPath), FPath) + Length(NormalDir(orignalPath)), length(FPath));

    StrLcopy(@Info.rfiName, PChar(FName), sizeof(Info.rfiName));
    StrLcopy(@Info.rfiPath, PChar(FPath), sizeof(Info.rfiPath));
    Info.rfiModifiedDate := FileDate(FileName);
    Info.rfiSize := Source.Size;

    result := Add(Info, Source, IAlgo);
  finally
    Source.Free;
  end;
end;

function TSystemCFileStruct.Add(const RealFileInfo: TRealFileInfo; const input: TStream; const IAlgo: integer): integer;
var Algo : TAlgoPlugins;
    Dest : TMultipleStream;
    Info : PFileInformation;
    BWritten,
    Index: Integer;
begin
  GetMem(Info, sizeof(Info^));
  try
    FillChar(Info^, sizeof(Info^), #0);

    // Fichier Reel
    Move(RealFileInfo, Info.fiRealFileInfo, sizeof(TRealFileInfo));

    // Fichier Compress
    if(IAlgo > -1) then Algo := FAlgo[IAlgo] else begin
      Index := FAlgo.IndexOfBest(input, cstQMax4Test);
      if(index < 0) then raise ESystemCFileStruct.Create(rsAlgoNotFound);
      Algo := FAlgo[Index];
    end;
    Info.fiCompressInfo.cfSerialAlgo := Algo.PluginsInformation.iSerial;

    AskOneDisk(FStartOfList.pNumDisk);
    FCurrentStream.Seek(FStartOfList.pPosInFile, soFromBeginning);

    Dest := TMultipleStream.Create(FCurrentStream, FCurrentMaxSize, OnWriteStream);
    try
      Algo.OnProgression := FOnProgress;
      Algo.Compression(BWritten, RealFileInfo.rfiSize, Input, Dest);
      Info.fiCompressInfo.cfPackedSize := (Dest as TMultipleStream).BytesWritten;
      Info.fiCompressInfo.cfCRC32      := (Dest as TMultipleStream).CRC32;
      Info.fiCompressInfo.cfErase      := false;
    finally
      Dest.Free;
    end;

    // Emplacement (aprs copie dans le fichier)
    Info.fiBlocAddress      := FStartOfList;
    FStartOfList.pNumDisk   := FCurrentDisk;
    FStartOfList.pPosInFile := FCurrentStream.Position;

    FNeedFlush := true;
    result     := FList.Add(Info);
  except
    FreeMem(Info);
    raise;
  end;
end;

procedure TSystemCFileStruct.AskNewDisk;
var MS: LongInt;
begin
  if(not Assigned(FOnAskNewDisk)) then
    raise ESystemCFileStruct.Create(rsNoAssignedEvent);

  Inc(FLastDisk);
  FCurrentDisk := FLastDisk;
  MS := CurrentMaxSize;
  if(not FOnAskNewDisk(FLastDisk, FCurrentStream, MS)) then raise ESystemCFileStruct.Create(rsAbort);
  CurrentMaxSize := MS;
  CreateHeadFile();
end;

procedure TSystemCFileStruct.AskOneDisk(const DiskNum: Byte);
var MS: LongInt;
begin
  if(DiskNum <> FCurrentDisk) then begin
    if(not Assigned(FOnAskOneDisk)) then
      raise ESystemCFileStruct.Create(rsNoAssignedEvent);

    MS := CurrentMaxSize;
    if(not FOnAskOneDisk(DiskNum, FCurrentStream, MS)) then raise ESystemCFileStruct.Create(rsAbort);
    CurrentMaxSize := MS;

    CheckFile();
  end;
end;

procedure TSystemCFileStruct.CheckFile;
var Buff: TIdentFile;
begin
  FCurrentStream.Seek(0, soFromBeginning);
  FCurrentStream.Read(Buff, SizeOf(Buff));
  if(Buff <> cstIdentFile) then raise ESystemCFileStruct.Create(rsBadFileFormat);
  FCurrentStream.Read(FCurrentDisk, SizeOf(FCurrentDisk));
end;

constructor TSystemCFileStruct.Create(const Archive: TStream; const MaxSize: LongInt; const List: TAlgoPluginsList; const OnNew: TOnAskNewDisk; const OnOne: TOnAskOneDisk);
begin
  FNeedFlush      := false;
  FOnAskNewDisk   := OnNew;
  FOnAskOneDisk   := OnOne;
  FOnProgress     := nil;
  FList           := TList.Create();
  FAlgo           := List;
  FCurrentStream  := Archive;
  CurrentMaxSize  := MaxSize;

  if(Archive.Size > 0) then begin
    CheckFile();
    FLastDisk := FCurrentDisk;
    LoadFileList();
  end else begin
    FCurrentDisk := 1;
    FLastDisk    := FCurrentDisk;
    CreateHeadFile();
    FStartOfList.pNumDisk   := FCurrentDisk;
    FStartOfList.pPosInFile := FCurrentStream.Position;
    Flush();
  end;
end;

procedure TSystemCFileStruct.CreateHeadFile;
begin
  FCurrentStream.Seek(0, soFromBeginning);
  FCurrentStream.Write(cstIdentFile, Sizeof(cstIdentFile));
  FCurrentStream.Write(FCurrentDisk, Sizeof(FCurrentDisk));
end;

procedure TSystemCFileStruct.Delete(index: Integer);
var I      : Integer;
    LastPos: Integer;
begin
  if(not IsMultiDisk) then begin
    FOnProgress(Format(rsDelete, [index]), 0, 1);
    LastPos := Items[Index].fiBlocAddress.pPosInFile;
    for i := Index + 1 to Count - 1 do begin // Pour tous les elments se trouvant aprs celui que l'on veut supprimer
      MoveStreamData(FCurrentStream, Items[I].fiBlocAddress.pPosInFile, LastPos, Items[I].fiCompressInfo.cfPackedSize);
      Items[I].fiBlocAddress.pPosInFile := LastPos;
      Inc(LastPos, Items[I].fiCompressInfo.cfPackedSize);
      FOnProgress(Format(rsDelete, [index]), I - Index, Count - Index);
    end;
    FStartOfList.pPosInFile := LastPos;
    FCurrentStream.Size     := LastPos;
    FreeMem(Items[Index]);
    FList.Delete(Index);
    FOnProgress(Format(rsDelete, [index]), 1, 1);
  end else begin
    Items[Index].fiCompressInfo.cfErase := true;
  end;
  FNeedFlush := true;
end;

destructor TSystemCFileStruct.Destroy;
var I: Integer;
begin
  if(FNeedFlush) then Flush();
  for I := 0 to Count - 1 do FreeMem(Items[I]);
  FList.Free;
  FAlgo := nil;
end;

procedure TSystemCFileStruct.Extract(Index: Integer; output: TStream);
var Algo: TAlgoPlugins;
    MSt : TMultipleStream;
begin
  Algo := FAlgo.Items[FAlgo.IndexOf(Items[Index].fiCompressInfo.cfSerialAlgo)];

  AskOneDisk(Items[Index].fiBlocAddress.pNumDisk);
  FCurrentStream.Seek(Items[Index].fiBlocAddress.pPosInFile, soFromBeginning);

  MSt := TMultipleStream.Create(FCurrentStream, FCurrentMaxSize, OnReadStream);
  try
    Algo.OnProgression := FOnProgress;
    Algo.Decompression(Items[Index].fiRealFileInfo.rfiSize, Items[Index].fiCompressInfo.cfPackedSize, MSt, output);
    if(Items[Index].fiCompressInfo.cfCRC32 <> MSt.CRC32) then
      raise ESystemCFileStruct.CreateFmt(rsBadCRC, [Trim(String(Items[Index].fiRealFileInfo.rfiName)),
                                                    IntToHex(MSt.CRC32, 8),
                                                    IntToHex(Items[Index].fiCompressInfo.cfCRC32, 8)]);
  finally
    MSt.Free;
  end;
end;

procedure TSystemCFileStruct.Extract(Index: Integer; path: TFileName);
var FileStream: TFileStream;
    FileName  : TFileName;
begin
  FileName   := path + Trim(String(Items[Index].fiRealFileInfo.rfiPath)) + Trim(String(Items[Index].fiRealFileInfo.rfiName));
  MakeAllDirs(ExtractFilePath(FileName));
  FileStream := TFileStream.Create(FileName, fmCreate);
  try
    Extract(Index, FileStream);
  finally
    FileStream.Free;
  end;
  SetFileDate(FileName, Items[Index].fiRealFileInfo.rfiModifiedDate);
end;

procedure TSystemCFileStruct.Flush;
var i  : integer;
    MSt: TMultipleStream;
    CRC: LongWord;
begin
  AskOneDisk(FStartOfList.pNumDisk);
  FCurrentStream.Seek(FStartOfList.pPosInFile, soFromBeginning);

  MSt := TMultipleStream.Create(FCurrentStream, FCurrentMaxSize, OnWriteStream);
  try
    if(Assigned(FOnProgress)) then OnProgress(rsFlush, 0, 1);
    for i := 0 to Count - 1 do begin
      MSt.Write(Items[I]^, SizeOf(TFileInformation));
      if(Assigned(FOnProgress)) then OnProgress(rsFlush, I + 1, Count);
    end;
    CRC := MSt.CRC32;
  finally
    MSt.Free;
  end;
  if((FCurrentStream.Size + cstMinSize) > CurrentMaxSize) then AskNewDisk;
  I := Count;
  FCurrentStream.Write(cstIdentFin, sizeof(TIdentFile));
  FCurrentStream.Write(FStartOfList, sizeof(TPosition));
  FCurrentStream.Write(I, sizeof(Integer));
  FCurrentStream.Write(CRC, sizeof(LongWord));

  FCurrentStream.Size := FCurrentStream.Position;
end;

function TSystemCFileStruct.GetCount: Integer;
begin
  result := FList.Count;
end;

function TSystemCFileStruct.GetItems(Index: Integer): PFileInformation;
begin
  result := PFileInformation(FList[Index]);
end;

function TSystemCFileStruct.IndexOf(FileName: TFileName; const index: integer): integer;
var I, Cnt: Integer;
begin
  Cnt    := 1;
  Result := -1;
  if(index < 1) then Exit;
  for i := 0 to Count - 1 do
    if(UpperCase(Trim(Items[I].fiRealFileInfo.rfiName)) = UpperCase(FileName)) then begin
      if(Cnt = Index) then begin
        result := I;
        break;
      end;
      Inc(Cnt);
    end;
end;

function TSystemCFileStruct.IsMultiDisk: boolean;
begin
  result := FLastDisk > 1;
end;

procedure TSystemCFileStruct.LoadFileList;
var BuffIdent: TIdentFile;
    Nombre, I: Integer;
    CRC      : LongWord;
    PInfo    : PFileInformation;
    MSt      : TMultipleStream;
begin
  FCurrentStream.Seek(- (sizeof(TIdentFile) + sizeof(TPosition) + sizeof(Integer) + sizeof(LongWord)), soFromEnd);
  FCurrentStream.Read(BuffIdent, sizeof(TIdentFile));
  if(BuffIdent <> cstIdentFin) then raise ESystemCFileStruct.Create(rsFileNotEnd);
  FCurrentStream.Read(FStartOfList, sizeof(TPosition));
  FCurrentStream.Read(Nombre, sizeof(Integer));
  FCurrentStream.Read(CRC, sizeof(LongWord));

  AskOneDisk(FStartOfList.pNumDisk);
  FCurrentStream.Seek(FStartOfList.pPosInFile, soFromBeginning);

  if(Assigned(FOnProgress)) then OnProgress(rsLoad, 0, Nombre);
  MSt := TMultipleStream.Create(FCurrentStream, FCurrentMaxSize, OnReadStream);
  try
    for i := 0 to Nombre - 1 do begin
      GetMem(PInfo, sizeof(TFileInformation));
      try
        MSt.Read(PInfo^, sizeof(TFileInformation));
      except
        FreeMem(PInfo);
        PInfo := nil;
      end;
      if(Assigned(PInfo)) then FList.Add(PInfo);
      if(Assigned(FOnProgress)) then OnProgress(rsLoad, I, Nombre);
    end;
    if(CRC <> MSt.CRC32) then raise ESystemCFileStruct.CreateFmt(rsBadCRC, [rsFileInfos]);
  finally
    MSt.Free;
  end;
end;

function TSystemCFileStruct.OnReadStream(var NewStream: TStream;
  var NewSpace: LongInt): boolean;
begin
  result      := (FCurrentDisk < FStartOfList.pNumDisk);
  if(result) then begin
    AskOneDisk(FCurrentDisk + 1);
    NewStream := FCurrentStream;
    NewSpace  := CurrentMaxSize;
  end;
end;

function TSystemCFileStruct.OnWriteStream(var NewStream: TStream;
  var NewSpace: LongInt): boolean;
begin
  result     := true;
  AskNewDisk();
  NewStream := FCurrentStream;
  NewSpace  := CurrentMaxSize;
end;

procedure TSystemCFileStruct.SetCurrentMaxSize(const Value: LongInt);
begin
  if(FCurrentMaxSize <> Value) then begin
    if(Value < cstMinSize) then
      raise ESystemCFileStruct.CreateFmt(rsMinSize, [cstMinSize]);
    FCurrentMaxSize := Value;
  end;
end;

procedure TSystemCFileStruct.SetOnAskNewDisk(const Value: TOnAskNewDisk);
begin
  FOnAskNewDisk := Value;
end;

procedure TSystemCFileStruct.SetOnAskOneDisk(const Value: TOnAskOneDisk);
begin
  FOnAskOneDisk := Value;
end;

procedure TSystemCFileStruct.SetOnProgress(const Value: TOnProgression);
begin
  FOnProgress := Value;
end;

end.
