(* GNU-GPL
 * System Compression - Logiciel de Compression/Decompression
 * Copyright (c) 2002 Ulrich Van Den Hekke (ulrich.vdh@free.fr)
 * 24 rue Jean Jaures
 * 59560 Comines
 *
 * Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon
 * les termes de la Licence Publique Gnrale GNU publie par la
 * Free Software Foundation (version 2 ou bien toute autre version ultrieure
 * choisie par vous).
 *
 * Ce programme est distribu car potentiellement utile, mais
 * SANS AUCUNE GARANTIE, ni explicite ni implicite, y compris les garanties de
 * commercialisation ou d'adaptation dans un but spcifique. Reportez-vous  la
 * Licence Publique Gnrale GNU pour plus de dtails.
 *
 * Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme
 * temps que ce programme ; si ce n'est pas le cas, crivez  la
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis.
 *
 *)

(******************************************************************************)
(* Unit contenant la fentre wizard permettant la cration de l'archive      *)
(* Version : 1.0.0.0                                                          *)
(* Author : Ulrich Van Den Hekke                                              *)
(* Create day : 02/09/2002                                                    *)
(******************************************************************************)

unit scWizard;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  AboutDlg, StdCtrls, ExtCtrls, ComCtrls, Buttons, scFileStruct,
  syswizconst, AlgoList, BrowseDr, Common, Menus, XPMan;
  
type
  EWizardForm = class(Exception);
  TWizardForm = class(TForm)
    PanelButton: TPanel;
    MainNotebook: TNotebook;
    ButtonAbout: TButton;
    AboutDlg: TAboutDlg;
    ButtonCancel: TButton;
    ButtonPrevious: TButton;
    ButtonNext: TButton;
    BigImage: TImage;
    BevelButton: TBevel;
    LabelTitre: TLabel;
    LabelTexte: TLabel;
    BevelTitre: TBevel;
    NotebookWizard: TNotebook;
    SmallImage: TImage;
    LabelQuestion: TLabel;
    LabelIndication: TLabel;
    LabelArchiveName: TLabel;
    BevelInfoArchive: TBevel;
    MemoInformation: TMemo;
    StaticArchiveName: TStaticText;
    OpenDialogArchive: TOpenDialog;
    LabelOperation: TLabel;
    RadioGroupOperation: TRadioGroup;
    AddListView: TListView;
    LabelList: TLabel;
    ButtonAdd: TButton;
    ButtonDelete: TButton;
    LabelDest: TLabel;
    EditDestDir: TEdit;
    ButtonBrownsDest: TButton;
    BrowseDirectoryDestDlg: TBrowseDirectoryDlg;
    LabelDeleteParent: TLabel;
    PanelArchiveList: TPanel;
    ArchiveListView: TListView;
    ButtonDeselect: TButton;
    ButtonSelect: TButton;
    LabelArchiveTitre: TLabel;
    LabelExtractParent: TLabel;
    LabelProgressFileName: TLabel;
    ProgressBarFile: TProgressBar;
    LabelProgressTotal: TLabel;
    ProgressBarTotal: TProgressBar;
    AnimateProgress: TAnimate;
    TimerProcessLaunch: TTimer;
    LabelProgressDisk: TLabel;
    ProgressBarDisque: TProgressBar;
    ButtonParcourirArchive: TSpeedButton;
    ButtonOldArchive: TSpeedButton;
    OldArchiveMenu: TPopupMenu;
    XPManifest1: TXPManifest;
    LabelRatio: TLabel;
    procedure ButtonAboutClick(Sender: TObject);
    procedure ButtonNextClick(Sender: TObject);
    procedure ButtonPreviousClick(Sender: TObject);
    procedure ButtonParcourirArchiveClick(Sender: TObject);
    procedure ButtonCancelClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ButtonAddClick(Sender: TObject);
    procedure AddListViewData(Sender: TObject; Item: TListItem);
    procedure ButtonDeleteClick(Sender: TObject);
    procedure ButtonSelectClick(Sender: TObject);
    procedure ButtonBrownsDestClick(Sender: TObject);
    procedure TimerProcessLaunchTimer(Sender: TObject);
    procedure ArchiveListViewData(Sender: TObject; Item: TListItem);
    procedure ButtonOldArchiveClick(Sender: TObject);
    procedure OldArchiveChooseClick(Sender: TObject);
  private
    FPluginsList : TAlgoPluginsList;

    FArchiveName : TFileName;
    FCurrentFile : TFileStream;
    FArchive     : TSystemCFileStruct;

    FAddList     : TList;
    FSelectList  : array of integer;

    FProgressFile: TCurrentFileInformation;
    FAddStream   : TStream;
    FChoosenAlgo : Integer;
    FOldProgress : byte;

    FAbort,
    FProgress    : boolean;
    { Dclarations prives }
    procedure AddDirectory2Archive(Parent, Directory: TFileName; Algo: Integer; recurse: boolean);
    procedure AddAFile2Archive(Info: TRealFileInfo; Algo: integer);

    procedure CreateArchive();
    procedure CloseArchive();
    procedure ProcessusTakeoff();
    procedure WaitNewDisk();
    procedure WaitOneDisk();
    procedure ChooseAlgoOnStream();
    procedure CalcRatio();

    function Progress(const Algo: string; const Position, Max: LongWord): boolean;
    function ProgressChoose(const Algo: string; const Position, Max: LongWord): boolean;

    procedure NextPage(CurrPage, SuccPage: TWizardPages);
    procedure ButtonVisible();
    procedure SetArchiveName(const Value: TFileName);
    procedure SetProgressionOfFile(const Value: integer);
    procedure SetProgressionTotal(Position: integer; const Value: integer);

    function WhenAskNewDisk(const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean;
    function WhenAskOneDisk(const DiskNum: Byte; var Stream: TStream; var MaxSize: LongInt): boolean;
  public
    { Dclarations publiques }
    procedure UpdateFreeSpace(Max: boolean; Disque: string);

    property ProgressionOfFile: integer write SetProgressionOfFile;
    property ProgressionTotal[Position: integer]: integer write SetProgressionTotal;
    property ArchiveName: TFileName read FArchiveName write SetArchiveName;
  end;

var
  WizardForm: TWizardForm;

implementation

uses FileTools, PaintUtils, syscconst, syscutils, scProgressForm, scAddForm, Math, ShlObj, RegInfos;

{$R *.DFM}

procedure TWizardForm.ButtonAboutClick(Sender: TObject);
begin
  AboutDlg.ExecuteAbout2;
end;

procedure TWizardForm.NextPage(CurrPage, SuccPage: TWizardPages);
begin
  try
    case(CurrPage) of
      wpNone          : begin
        ArchiveName := '';
        if(ParamStr(1) <> EmptyStr) then begin
          if(length(ExtractFileExt(ParamStr(1))) <= 1) then
            ArchiveName := ChangeFileExt(ParamStr(1), '.' + OpenDialogArchive.DefaultExt)
          else
            ArchiveName := ParamStr(1);

          SuccPage := wpOperationChoice;
          NextPage(wpArchiveChoice, SuccPage);
          Exit;
        end;
      end;
      wpArchiveChoice : case(SuccPage) of
        wpNone : ButtonNext.Enabled := true;
        wpOperationChoice: begin
          CreateProgressDialogForm(rsLoadPlugins, CreateArchive);
          if(not (Assigned(FCurrentFile) and Assigned(FArchive) and Assigned(FPluginsList))) then
            raise EWizardForm.Create(rsGoBack);
          RadioGroupOperation.Items.Clear;
          RadioGroupOperation.Items.Add(rsOpChoice0);
          RadioGroupOperation.Items.Add(rsOpChoice1);
          if(not FArchive.IsMultiDisk) then
            RadioGroupOperation.Items.Add(rsOpChoice2);
          RadioGroupOperation.ItemIndex := 1;
          if(FArchive.Count = 0) then begin
            SuccPage := wpOperationAdd;
            NextPage(wpOperationChoice, SuccPage);
            RadioGroupOperation.ItemIndex := 0;
          end;
        end;
      end;
      wpOperationChoice : begin
        case(SuccPage) of
          wpArchiveChoice   : CreateProgressDialogForm(rsCloseArchive, CloseArchive);
          wpOperationExtract: begin
            PanelArchiveList.Parent     := LabelExtractParent.Parent;
            LabelArchiveTitre.Caption   := Format(rsLabelList, [rsExtractText]);
            ArchiveListView.Items.Count := FArchive.Count;
            ButtonSelect.Click;
          end;
          wpOperationDelete : begin
            PanelArchiveList.Parent     := LabelDeleteParent.Parent;
            LabelArchiveTitre.Caption   := Format(rsLabelList, [rsDeleteText]);
            ArchiveListView.Items.Count := FArchive.Count;
            ButtonDeselect.Click;
          end;
        end;
        if(SuccPage in [wpOperationExtract, wpOperationDelete]) then
          CreateProgressDialogForm(rsCalcRatio, CalcRatio);
      end;
      wpOperationAdd : case(SuccPage) of
        wpOperationChoice: begin
          if(FArchive.Count = 0) then begin
            SuccPage := wpArchiveChoice;
            NextPage(wpOperationChoice, SuccPage);
          end;
        end;
        wpOperationProgress : begin
          CreateProgressDialogForm(rsTakeoff, ProcessusTakeoff);
        end;
      end;
      wpOperationExtract: case(SuccPage) of
        wpOperationDest: begin
          EditDestDir.Text := ExtractFilePath(FArchiveName);
        end;
      end;
      wpOperationDest: case(SuccPage) of
        wpOperationProgress : begin
          EditDestDir.Text := ExpandFileName(EditDestDir.Text);
          CreateProgressDialogForm(rsTakeoff, ProcessusTakeoff);
        end;
      end;
      wpOperationDelete: case(SuccPage) of
        wpOperationProgress : begin
          CreateProgressDialogForm(rsTakeoff, ProcessusTakeoff);
        end;
      end;
    end;
    LabelQuestion.Caption    := cstTitlePage[SuccPage].Question;
    LabelIndication.Caption  := cstTitlePage[SuccPage].Sujet;
    NotebookWizard.PageIndex := Ord(SuccPage);
 except
   on E: Exception do begin
     MessageDlg(E.Message, mtError, [mbOk], 0);
   end;
 end;
end;

procedure TWizardForm.ButtonNextClick(Sender: TObject);
begin
  if(MainNotebook.PageIndex = Ord(mpBigBitmap)) then begin
    MainNotebook.PageIndex := Ord(mpSmallBitmap);
    NextPage(wpNone, wpArchiveChoice);
  end else
  case TWizardPages(NotebookWizard.PageIndex) of
    wpArchiveChoice: NextPage(wpArchiveChoice, wpOperationChoice);
    wpOperationChoice: case(RadioGroupOperation.ItemIndex) of
      cstAdd    : NextPage(wpOperationChoice, wpOperationAdd);
      cstExtract: NextPage(wpOperationChoice, wpOperationExtract);
      cstDelete : NextPage(wpOperationChoice, wpOperationDelete);
    end;
    wpOperationAdd    : NextPage(wpOperationAdd,     wpOperationProgress);
    wpOperationExtract: NextPage(wpOperationExtract, wpOperationDest);
    wpOperationDest   : NextPage(wpOperationDest,    wpOperationProgress);
    wpOperationDelete : NextPage(wpOperationDelete,  wpOperationProgress);
  end;
  ButtonVisible();
end;

procedure TWizardForm.ButtonPreviousClick(Sender: TObject);
begin
  case TWizardPages(NotebookWizard.PageIndex) of
    wpArchiveChoice: begin
      MainNotebook.PageIndex := Ord(mpBigBitmap);
      NextPage(wpArchiveChoice, wpNone);
    end;
    wpOperationChoice   : NextPage(wpOperationChoice, wpArchiveChoice);
    wpOperationAdd      : NextPage(wpOperationAdd, wpOperationChoice);
    wpOperationExtract  : NextPage(wpOperationExtract, wpOperationChoice);
    wpOperationDest     : NextPage(wpOperationDest, wpOperationExtract);
    wpOperationDelete   : NextPage(wpOperationDelete, wpOperationChoice);
    wpOperationProgress : ;
  end;
  ButtonVisible();
end;

procedure TWizardForm.ButtonVisible;
begin
  ButtonNext.Visible     := (NotebookWizard.PageIndex <> Ord(wpOperationProgress));
  ButtonPrevious.Visible := (MainNotebook.PageIndex <> Ord(mpBigBitmap)) and
                            (NotebookWizard.PageIndex <> Ord(wpOperationProgress));
end;

procedure TWizardForm.SetArchiveName(const Value: TFileName);
begin
  FArchiveName := Value;

  MemoInformation.Lines.Clear;
  ButtonNext.Enabled := FArchiveName <> '';

  if(ButtonNext.Enabled) then begin
    MemoInformation.Lines.Add(Format(rsMemoInfo1, [ExtractFileName(FArchiveName)]));
    if(FileExists(FArchiveName)) then begin
      MemoInformation.Lines.Add(Format(rsMemoInfo2, [rsYes]));
    end else begin
      MemoInformation.Lines.Add(Format(rsMemoInfo2, [rsNo]));
    end;
    StaticArchiveName.Caption := MinimizeFileName(FArchiveName,
      Round(StaticArchiveName.Width / (Canvas.TextWidth(FArchiveName) / Length(FArchiveName))));
  end else
    StaticArchiveName.Caption := rsNone;
  Caption := Format(rsTitle, [ExtractFileName(FArchiveName)]);
end;

procedure TWizardForm.ButtonParcourirArchiveClick(Sender: TObject);
begin
  if(OpenDialogArchive.Execute) then begin
    ArchiveName := OpenDialogArchive.FileName;
  end;
end;

procedure TWizardForm.CreateArchive;
var MRUIndex: Integer;
begin
  FCurrentFile := nil; FArchive := nil; FPluginsList := nil;
  try
    FPluginsList := TAlgoPluginsList.Create(RegInfosOption.PluginsDir);

    ProgressForm.PanelText.Caption := rsOpenArchive;

    MRUIndex := RegInfosOption.MRUList.IndexOf(ArchiveName);
    if(MRUIndex <> -1) then RegInfosOption.MRUList.Delete(MRUIndex);

    if(FileExists(ArchiveName)) then
      FCurrentFile := TFileStream.Create(ArchiveName, fmOpenReadWrite or fmShareDenyWrite)
    else
      FCurrentFile := TFileStream.Create(ArchiveName, fmCreate or fmShareDenyWrite);

    FArchive :=  TSystemCFileStruct.Create(FCurrentFile,
                                           CalcFileSize(FCurrentFile, ArchiveName),
                                           FPluginsList, WhenAskNewDisk, WhenAskOneDisk);
    FArchive.OnProgress := Progress;
  except
    on E: Exception do begin
      FPluginsList.Free;
      FArchive.Free;
      FCurrentFile.Free;
      MessageDlg(E.Message, mtError, [mbOk], 0);
    end;
  end;
end;

procedure TWizardForm.CloseArchive;
begin
  if(Assigned(FArchive)) then begin
    RegInfosOption.MRUList.Add(ArchiveName);
    SHAddToRecentDocs(SHARD_PATH, PChar(ArchiveName));
  end;
  FArchive.Free;
  FCurrentFile.Free;
  ProgressForm.PanelText.Caption := rsClosePlugins;
  FPluginsList.Free;
  FCurrentFile := nil; FArchive := nil; FPluginsList := nil;
end;

procedure TWizardForm.ButtonCancelClick(Sender: TObject);
begin
  Close;
end;

procedure TWizardForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  CanClose := not FProgress;
  if(not CanClose) then begin
    if(MessageDlg(rsQueryClose, mtConfirmation, mbOKCancel, 0) = mrOk) then
      FAbort := true;
  end else
    FAbort := not FAbort;
end;

procedure TWizardForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if(MainNotebook.PageIndex = Ord(mpSmallBitmap)) then begin
    Action := caNone;

    CreateProgressDialogForm(rsCloseArchive, CloseArchive);
    ButtonPrevious.Visible := false;
    ButtonNext.Visible     := false;
    ButtonCancel.Caption   := rsTerminate;

    if(FAbort) then begin
      LabelTitre.Caption     := rsAbort;
      LabelTexte.Caption     := rsAbortText;
    end else begin
      LabelTitre.Caption     := rsFinish;
      LabelTexte.Caption     := rsFinishText;
    end;

    MainNotebook.PageIndex := Ord(mpBigBitmap);
  end;
end;

procedure TWizardForm.FormCreate(Sender: TObject);
begin
  NotebookWizard.ControlStyle := NotebookWizard.ControlStyle - [csParentBackground];
  MainNotebook.ControlStyle := NotebookWizard.ControlStyle - [csParentBackground];

  FAddList := TList.Create;

  FAbort                 := false;
  FProgress              := false;
  MainNotebook.PageIndex := Ord(mpBigBitmap);
  LabelTitre.Caption     := rsWelcom;
  LabelTexte.Caption     := rsWelcomText;

  if((RegInfosOption.WizardPos.x <> -1) and (RegInfosOption.WizardPos.y <> -1)) then
    SetBounds(RegInfosOption.WizardPos.x, RegInfosOption.WizardPos.y, Width, Height);
end;

procedure TWizardForm.FormDestroy(Sender: TObject);
var i: integer;
    P: PAddingFile;
begin
  RegInfosOption.WizardPos.x := Left;
  RegInfosOption.WizardPos.y := Top;

  for i := 0 to FAddList.Count - 1 do begin
    P := PAddingFile(FAddList[I]);
    case(P.afMode) of
      atFile     : FreeMem(P.afFileName);
      atDirectory: FreeMem(P.afDirName);
      atWebAddr  : begin
        FreeMem(P.afAddrName);
        if(Assigned(P.afProxyName)) then FreeMem(P.afProxyName);
      end;
    end;
    FreeMem(P);
  end;
  FAddList.Free;
end;

procedure TWizardForm.ButtonAddClick(Sender: TObject);
var ListP: TList;
    I    : Integer;
begin
  ListP := GetAddFile(FPluginsList);
  for I := 0 to ListP.Count - 1 do
    FAddList.Add(ListP[I]);
  ListP.Free;
  AddListView.Items.Count := FAddList.Count;
end;

procedure TWizardForm.AddListViewData(Sender: TObject; Item: TListItem);
var ShowItem: PAddingFile;
begin
  if((not (Assigned(FAddList) and Assigned(FPluginsList))) or (Item.Index >= FAddList.Count)) then Exit;
  ShowItem := PAddingFile(FAddList[Item.Index]);
  case(ShowItem.afMode) of
    atFile     : begin
      Item.Caption := ShowItem.afFileName;
      Item.SubItems.Add(rsNo);
      Item.SubItems.Add(EmptyStr);
      Item.SubItems.Add(EmptyStr);
    end;
    atDirectory: begin
      Item.Caption := ShowItem.afDirName;
      if(ShowItem.afRecurse) then
        Item.SubItems.Add(rsYes)
      else
        Item.SubItems.Add(rsNo);
      Item.SubItems.Add(EmptyStr);
      Item.SubItems.Add(EmptyStr);
    end;
    atWebAddr  : begin
      Item.Caption := ShowItem.afAddrName;
      if(ShowItem.afProxyName <> EmptyStr) then
        Item.SubItems.Add(rsYes)
      else
        Item.SubItems.Add(rsNo);
      Item.SubItems.Add(ShowItem.afProxyName);
      Item.SubItems.Add(IntToStr(ShowItem.afProxyPort));
    end;
  end;
  try
    if(ShowItem.afAlgoIndex >= 0) then
      Item.SubItems.Add(Trim(FPluginsList[ShowItem.afAlgoIndex].PluginsInformation.iName))
    else
      Item.SubItems.Add(cstAutoPlugin);
  except
      Item.SubItems.Add(EmptyStr);
  end;
end;

procedure TWizardForm.ButtonDeleteClick(Sender: TObject);
begin
  if(AddListView.Selected <> nil) then begin
    FreeMem(FAddList[AddListView.Selected.Index]);
    FAddList.Delete(AddListView.Selected.Index);
    AddListView.Items.Count := FAddList.Count;
    AddListView.Repaint;
  end;
end;

procedure TWizardForm.ButtonSelectClick(Sender: TObject);
var i: integer;
begin
  ArchiveListView.Items.BeginUpdate();
  try
    for i := 0 to ArchiveListView.Items.Count - 1 do
      ArchiveListView.Items[I].Selected := Boolean((Sender as TButton).Tag);
  finally
    ArchiveListView.Items.EndUpdate();
  end;
end;

procedure TWizardForm.ButtonBrownsDestClick(Sender: TObject);
begin
  BrowseDirectoryDestDlg.Selection := EditDestDir.Text;
  if(BrowseDirectoryDestDlg.Execute) then
    EditDestDir.Text := BrowseDirectoryDestDlg.Selection;
end;

procedure TWizardForm.ArchiveListViewData(Sender: TObject; Item: TListItem);
var IndexPlugin: Integer;
begin
  if((not (Assigned(FArchive) and Assigned(FPluginsList))) or (Item.Index >= FArchive.Count)) then Exit;
  with(FArchive[Item.Index]^) do begin
    Item.Caption := Trim(fiRealFileInfo.rfiName);
    Item.SubItems.Add(DateToStr(fiRealFileInfo.rfiModifiedDate));
    Item.SubItems.Add(SizeToStr(fiRealFileInfo.rfiSize));
    if(fiRealFileInfo.rfiSize <> 0) then
      Item.SubItems.Add(Format('%d%%', [100 - Round((fiCompressInfo.cfPackedSize / fiRealFileInfo.rfiSize)  * 100)]))
    else
      Item.SubItems.Add('-');
    Item.SubItems.Add(SizeToStr(fiCompressInfo.cfPackedSize));
    Item.SubItems.Add(IntToHex(fiCompressInfo.cfCRC32, 8));
    IndexPlugin := FPluginsList.IndexOf(fiCompressInfo.cfSerialAlgo);
    if(IndexPlugin >= 0) then
      Item.SubItems.Add(Trim(FPluginsList[IndexPlugin].PluginsInformation.iName))
    else
      Item.SubItems.Add(rsUnsuported);
    Item.SubItems.Add(IntToStr(fiBlocAddress.pNumDisk));
  end;
end;

procedure TWizardForm.SetProgressionOfFile(const Value: integer);
begin
  ProgressBarFile.Position      := Value;
  ProgressBarFile.Max           := FProgressFile.TailleOfFile;
  LabelProgressFileName.Caption := Format(rsProgressFile, [FProgressFile.FileName, FProgressFile.Algo,
                                     SizeToStr(Value), SizeToStr(FProgressFile.TailleOfFile)]);
end;

procedure TWizardForm.SetProgressionTotal(Position: integer; const Value: integer);
begin
  ProgressBarTotal.Max      := Value;
  ProgressBarTotal.Position := Position;
  LabelProgressTotal.Caption:= Format(rsProgressTotal, [Position, Value]);
end;

procedure TWizardForm.UpdateFreeSpace(Max: boolean; Disque: string);
begin
  if(Max) then begin
    LabelProgressDisk.Caption  := '';
    ProgressBarDisque.Max      := Min(DiskFree(Disque), High(Integer));
  end else begin
    LabelProgressDisk.Caption  := Format(rsProgressFree, [SizeToStr(DiskFree(Disque)), FArchive.CurrentDisk, FArchive.NumberOfDisk]);
    ProgressBarDisque.Position := Min(DiskFree(Disque), High(Integer));
  end;
end;

procedure TWizardForm.ProcessusTakeoff;
var Item: TListItem;
    I   : integer;
begin
  LabelProgressTotal.Caption    := '';
  LabelProgressFileName.Caption := '';
  LabelProgressDisk.Caption     := '';

  if(RadioGroupOperation.ItemIndex = cstDelete) then begin
    LabelProgressDisk.Visible := false;
    ProgressBarDisque.Visible := false;
  end;

  FAbort    := false;
  FProgress := true;

  SetLength(FSelectList, ArchiveListView.SelCount);
  Item := ArchiveListView.Selected; I := 0;
  while(Item <> nil) do begin
    FSelectList[I] := Item.Index;
    Item := ArchiveListView.GetNextItem(Item, sdBelow, [isSelected]);
    Inc(I);
  end;

  TimerProcessLaunch.Enabled    := true;
end;

procedure TWizardForm.TimerProcessLaunchTimer(Sender: TObject);
var Operation, I: integer;
    FileInfo    : PAddingFile;
    RealFileInfo: TRealFileInfo;
begin
  TimerProcessLaunch.Enabled := false;
  AnimateProgress.Active     := true;
  Operation                  := RadioGroupOperation.ItemIndex;

  try
    if(Operation = cstAdd) then begin
      UpdateFreeSpace(true, FArchiveName);
      for I := 0 to FAddList.Count - 1 do begin
        ProgressionTotal[I] := FAddList.Count;
        FileInfo            := FAddList[I];
        case(FileInfo.afMode) of
          atFile : begin
            FAddStream := TFileStream.Create(FileInfo.afFileName, fmOpenRead or fmShareDenyWrite);

            FillChar(RealFileInfo, Sizeof(RealFileInfo), #0);
            StrPLCopy(@RealFileInfo.rfiName, ExtractFileName(FileInfo.afFileName), sizeof(RealFileInfo.rfiName));
//            StrPLCopy(@RealFileInfo.rfiPath, EmptyStr,                             sizeof(RealFileInfo.rfiPath));
            RealFileInfo.rfiModifiedDate := FileDate(FileInfo.afFileName);
            RealFileInfo.rfiSize         := FAddStream.Size;
            AddAFile2Archive(RealFileInfo, FileInfo.afAlgoIndex);
          end;
          atDirectory : begin
            AddDirectory2Archive(FileInfo.afDirName, FileInfo.afDirName, FileInfo.afAlgoIndex, FileInfo.afRecurse);
          end;
{ TODO 2 -oUlrich -cProgrammation : Ajout du support internet }
          else MessageDlg('Not Implemented', mtWarning, [mbOk], 0);
        end;

        if(FAbort) then raise EAbort.Create('Abort');
      end;
      ProgressionTotal[FAddList.Count] := FAddList.Count;
    end else begin
      if(Operation = cstExtract) then UpdateFreeSpace(true, EditDestDir.Text) else UpdateFreeSpace(true, FArchiveName);
      for I := 0 to High(FSelectList) do begin
        ProgressionTotal[I]    := High(FSelectList) + 1;
        if(Operation = cstDelete) then Dec(FSelectList[I], I);
        FProgressFile.FileName := Trim(FArchive[FSelectList[I]].fiRealFileInfo.rfiName);
        try
          if(Operation = cstDelete) then
            FArchive.Delete(FSelectList[I])
          else begin
            if(FArchive.AlgoList.IndexOf(FArchive[FSelectList[I]].fiCompressInfo.cfSerialAlgo) < 0) then
              MessageDlg(rsCantExtUnsupp, mtInformation, [mbOK], 0)
            else
              FArchive.Extract(FSelectList[I], EditDestDir.Text);
          end;
        except
          on E: Exception do if(not FAbort) then
            if(MessageDlg(E.Message, mtError, [mbIgnore, mbAbort], 0) = mrAbort) then
              raise EAbort.Create('Abort');
        end;
        if(FAbort) then raise EAbort.Create('Abort');
      end;
      ProgressionTotal[High(FSelectList) + 1] := High(FSelectList) + 1;
    end;

    FAbort := true;
  except
    FAddStream.Free;
    FAbort := false;
  end;

  SetLength(FSelectList, 0);
  FProgress := false;
  Close();
end;

function TWizardForm.Progress(const Algo: string; const Position, Max: LongWord): boolean;
var Percent  : byte;
begin
  result   := not FAbort;

  Percent  := Round((Position / Max) * 100);
  if(Percent = FOldProgress) then Exit;
  FOldProgress := Percent;

  FProgressFile.Algo         := Trim(Algo);
  FProgressFile.TailleOfFile := Max;
  ProgressionOfFile          := Position;

  if(RadioGroupOperation.ItemIndex = cstExtract) then
    UpdateFreeSpace(false, EditDestDir.Text)
  else
    UpdateFreeSpace(false, FArchiveName);
  Application.ProcessMessages;
end;

function TWizardForm.WhenAskNewDisk(const DiskNum: Byte;
  var Stream: TStream; var MaxSize: Integer): boolean;
begin
  result := true;

  FCurrentFile.Free;

  CreateProgressDialogForm(Format(rsInsertNewDisk, [UpCase(ArchiveName[1])]), WaitNewDisk);

  UpdateFreeSpace(true, FArchiveName);

  ArchiveName := GenerateFileName(ArchiveName, DiskNum);

  FCurrentFile := TFileStream.Create(ArchiveName, fmCreate or fmShareDenyWrite);
  MaxSize      := CalcFileSize(FCurrentFile, ArchiveName);
  Stream       := FCurrentFile;
end;

function TWizardForm.WhenAskOneDisk(const DiskNum: Byte;
  var Stream: TStream; var MaxSize: Integer): boolean;
begin
  result := true;

  FCurrentFile.Free;
  ArchiveName := GenerateFileName(ArchiveName, DiskNum);

  CreateProgressDialogForm(Format(rsInsertOneDisk, [DiskNum, ArchiveName[1]]), WaitOneDisk);

  FCurrentFile := TFileStream.Create(ArchiveName, fmOpenReadWrite or fmShareDenyWrite);
  MaxSize      := CalcFileSize(FCurrentFile, ArchiveName);
  Stream       := FCurrentFile;
end;

procedure TWizardForm.WaitNewDisk;
begin
  while(not (DiskInDrive(ArchiveName[1])) or (DiskFree(ArchiveName) <= cstMinSize)) do begin
    Sleep(2000);
    Application.ProcessMessages();
  end;
end;

procedure TWizardForm.WaitOneDisk;
begin
  while(not (DiskInDrive(ArchiveName[1])) or (not FileExists(ArchiveName))) do begin
    Sleep(2000);
    Application.ProcessMessages();
  end;
end;

procedure TWizardForm.ChooseAlgoOnStream;
var StCount: Integer;
begin
  StCount := Max(Round(FAddStream.Size * cstPercentTest), cstQMax4Test);
  FPluginsList.OnProgression := ProgressChoose;
  FChoosenAlgo := FPluginsList.IndexOfBest(FAddStream, StCount);
end;

function TWizardForm.ProgressChoose(const Algo: string; const Position, Max: LongWord): boolean;
var Percent  : byte;
begin
  result := true;
  Percent  := Round((Position / Max) * 255);
  if(Percent = FOldProgress) then Exit;
  FOldProgress := Percent;

  LabelProgressFileName.Caption := Format(rsTestAlgo, [Trim(Algo), FProgressFile.FileName]);
  Application.ProcessMessages();
end;

procedure TWizardForm.AddAFile2Archive(Info: TRealFileInfo; Algo: integer);
begin
  try
    FProgressFile.FileName       := Trim(Info.rfiName);
    FChoosenAlgo                 := Algo;
    if(FChoosenAlgo < 0) then ChooseAlgoOnStream;

    FArchive.Add(Info, FAddStream, FChoosenAlgo);
    FAddStream.Free;
    FAddStream := nil;
  except
    on E: Exception do if(not FAbort) then
      if(MessageDlg(E.Message, mtError, [mbIgnore, mbAbort], 0) = mrAbort) then
        raise EAbort.Create('Abort');
  end;
  Application.ProcessMessages;
end;

procedure TWizardForm.AddDirectory2Archive(Parent, Directory: TFileName; Algo: Integer; recurse: boolean);
var SRec: TSearchRec;
    res : Integer;
    RFI : TRealFileInfo;
begin
  FillChar(SRec, sizeof(SRec), #0);
  Directory := NormalDir(Directory);
  res := FindFirst(Directory + '*.*', faAnyFile, SRec);
  while((not FAbort) and (res = 0)) do begin
    if(not ((SRec.Attr and faVolumeID > 0) or (SRec.Attr and faDirectory > 0))) then begin
      FAddStream := TFileStream.Create(Directory + SRec.Name, fmOpenRead or fmShareDenyWrite);

      FillChar(RFI, Sizeof(RFI), #0);
      StrPLCopy(@RFI.rfiName, SRec.Name, sizeof(RFI.rfiName));
      StrPLCopy(@RFI.rfiPath, Copy(Directory, Pos(Parent, Directory) + Length(Parent), length(Directory)), sizeof(RFI.rfiPath));
      RFI.rfiModifiedDate := FileDate(Directory + SRec.Name);
      RFI.rfiSize         := FAddStream.Size;
      AddAFile2Archive(RFI, Algo);
    end;
    res := FindNext(SRec);
  end;
  FindClose(SRec);

  if(recurse) then begin
    res  := FindFirst(Directory + '*.*', faAnyFile, SRec);
    while((not FAbort) and (res = 0)) do begin
      if((SRec.Attr and faDirectory > 0) and (SRec.Name <> '.') and (SRec.Name <> '..')) then
        AddDirectory2Archive(Parent, Directory + SRec.Name, Algo, Recurse);
      res := FindNext(SRec);
    end;
    FindClose(SRec);
  end;
end;

procedure TWizardForm.ButtonOldArchiveClick(Sender: TObject);
var i, c: integer;
    Menu: TMenuItem;
begin
  c := -1;
  for i := OldArchiveMenu.Items.Count - 1 downto 0 do begin
    if(OldArchiveMenu.Items[I].Checked) then c := i;
    OldArchiveMenu.Items.Delete(I);
  end;
  for i := 0 to RegInfosOption.MRUList.Count - 1 do begin
    Menu := TMenuItem.Create(OldArchiveMenu.Items);
    Menu.Caption    := '&' + IntToStr(I) + ' - ' + ExtractFileName(RegInfosOption.MRUList[i]);
    Menu.Tag        := I;
    Menu.GroupIndex := 1;
    Menu.RadioItem  := true;
    Menu.OnClick    := OldArchiveChooseClick;
    if(c = i) then Menu.Checked := true;
    OldArchiveMenu.Items.Add(Menu);
  end;
  OldArchiveMenu.Popup(ButtonOldArchive.ClientOrigin.x, ButtonOldArchive.ClientOrigin.y + ButtonOldArchive.Height);
end;

procedure TWizardForm.OldArchiveChooseClick(Sender: TObject);
begin
  ArchiveName := RegInfosOption.MRUList[(Sender as TMenuItem).Tag];
  (Sender as TMenuItem).Checked := true;
end;

procedure TWizardForm.CalcRatio;
var I              : Integer;
    RatioC, RatioD : Int64;
begin
  Application.ProcessMessages;
  RatioC := 0; RatioD := 0;
  for I := 0 to FArchive.Count - 1 do begin
    RatioC := RatioC + FArchive[I].fiCompressInfo.cfPackedSize;
    RatioD := RatioD + FArchive[I].fiRealFileInfo.rfiSize;
  end;
  LabelRatio.Caption := Format(rsCompressRatio, [(RatioC / RatioD) * 100]);
end;

end.
